from multiprocessing import Process, Queue, Lock
from queue import Empty
from nfer.nfer import globcallbacks
from signal import signal, SIGTERM
import atexit
from aiohttp import web
import asyncio
import socketio
import logging
import os
import webbrowser


MIN_DURATION = 0
UPDATE_PERIOD_MS = 200
log = logging.getLogger('nfer')

# IPC queue for handling interval output
gui_queue = Queue(0)
connected = Lock()
gui_proc = None

# initial parameters for the gui
display_intervals = []
display_tooltips = []

############# set up AIOHTTP webserver #############
sio = socketio.AsyncServer(async_mode='aiohttp')
app = web.Application()
sio.attach(app)

# get the directories for serving files
current_directory = os.path.dirname(os.path.abspath(__file__))
static_directory = os.path.join(current_directory, 'static')
www_directory = os.path.join(current_directory, 'www')

# configure static files
app.add_routes([web.static('/static', static_directory), 
                web.static('/', www_directory, show_index=True)])

##################################################

@sio.on("init")
def handle_connection(sid, data):
    #print("received json: " + str(data))
    # set the connected variable so the gui method will return
    global connected
    connected.release()
    # return configuration
    return {'intervals':display_intervals, 'tooltips':display_tooltips}

def subscriber(interval):
    # try only outputting intervals with long enough duration
    if interval.end - interval.begin > MIN_DURATION:
        gui_queue.put({'name':interval.name, 'start':interval.begin, 'end':interval.end, 'data':interval.data})

async def emit_queue():
    period = UPDATE_PERIOD_MS / 1000
    batch = []
    # wait for a cancel signal from asyncio
    try:
        # loop forever (until cancelled)
        while True:
            # get all the current items in the queue (this could be problematic if they are coming very quickly...)
            while True:
                try:
                    interval = gui_queue.get_nowait()
                except Empty:
                    break
                else:
                    batch.append(interval)
                    # if this was the standard threading queue, we'd need to call task_done()

            # emit the batch
            await sio.emit('batch', batch)
            batch.clear()
            # enqueue the next round
            await asyncio.sleep(period)
    except asyncio.CancelledError:
        pass
    # finally:
        # consider adding a finally here to shut anything down we need


async def start_background_tasks(app):
    app['queue_emitter'] = asyncio.create_task(emit_queue())

async def cleanup_background_tasks(app):
    app['queue_emitter'].cancel()
    await app['queue_emitter']

def run_nfer_gui(intervals, tooltips, port=5000):
    # lots of globals in this - consider wrapping in an object
    global display_intervals, display_tooltips
    # , srv, loop

    # set the initial arguments
    display_intervals = intervals
    display_tooltips = tooltips


    # AIOHTTP has a clean way to add other asyncio tasks to the event loop
    # You just need to register the task with AIOHTTP to start and stop alongside it
    app.on_startup.append(start_background_tasks)
    app.on_cleanup.append(cleanup_background_tasks)
    web.run_app(app, port=port)


def shutdown_nfer_gui():
    global gui_proc
    log.debug('nfer gui process: called shutdown!')
    gui_proc.terminate()
    gui_proc.join()
    log.debug('nfer guid process: joined!')

def gui(intervals, tooltips=['args'], port=5000, wait=True, browser=True):
    global gui_proc
    # add callbacks for all the intervals to monitor
    # for name in intervals:
    #     if name not in callbacks:
    #         callbacks[name] = []
    #     callbacks[name].append(subscriber)
    globcallbacks.append(subscriber)
    # get the lock so we can wait in this process for the client to connect to the web server
    connected.acquire()

    # add a handler so we can tell when the main process exits
    atexit.register(shutdown_nfer_gui)

    # start the web server in a subprocess
    gui_proc = Process(name='nfer.gui', target=run_nfer_gui, kwargs={'intervals':intervals,'tooltips':tooltips,'port':port})
    gui_proc.start()

    # if we are to open a web browser window, do so
    webbrowser.open('http://localhost:%d/index.html' % port)

    # if we are to wait, pause execution until the client connects
    if wait:
        log.info('nfer console is waiting for a connection on http://localhost:%d' % port)
        connected.acquire()





def normalize_interval_name(name):
    # replace periods with octothorps
    return name.replace(".","#")

def denormalize_interval_name(name):
    # replace octothorps with periods
    return name.replace("#",".")


from queue import Queue, Empty
from threading import Thread, current_thread, main_thread, enumerate
import _nfer
import time
import logging
from nfer.rules import Rule
from nfer.normalize import normalize_interval_name, denormalize_interval_name

"""
This file contains the primary API for monitoring and reporting intervals
"""

log = logging.getLogger('nfer')

class Interval:
    def __init__(self, name, begin = None, end = None, data = {}):
        if isinstance(name, dict):
            self.name = name["name"]
            self.begin = name["begin"]
            self.end = name["end"]
            self.data = name
        else:
            self.name = name
            self.begin = begin
            self.end = end
            self.data = data
        # now try to fix things - try to do the right thing if data is missing
        if self.begin is None:
            self.begin = now()
        if self.end is None:
            self.end = self.begin

    # this gets data as if though it is an attribute
    def __getattr__(self, field):
        return self.data[field]

    def __str__(self):
        return f"Interval({self.name},{self.begin},{self.end},{str(self.data)})"

# threadsafe queue for handling interval input
input_queue = Queue(0)
callbacks = {}
globcallbacks = []

now = lambda: int(round(time.time() * 1000))

def report_event(name, data={}):
    if isinstance(name, str):
        ts = now()
        report_interval(Interval(name, ts, ts, data))

def report_interval(interval):
    if isinstance(interval, Interval):
        # send to nfer, which is running in a separate thread
        input_queue.put(interval)
        # also call any callbacks for this interval
        if interval.name in callbacks.keys():
            for callback in callbacks[interval.name]:
                callback(interval)
        # also call any global callbacks
        for callback in globcallbacks:
            callback(interval)

def nfer_daemon():
    """
    Monitor the input queue for reported intervals and pass them to the C backend.
    This function is meant to be run as a daemon thread and will not return.
    :return: Does not return
    """
    # get all the current items in the queue (this could be problematic if they are coming very quickly...)
    while True:
        try:
            interval = input_queue.get(timeout=1)
        except Empty:
            log.debug('timed out')
            if log.isEnabledFor(logging.DEBUG):
                for thread in enumerate():
                    log.debug('Running thread: %s' % thread.name)
            continue

        # normalize the name before passing to nfer
        norm_name = normalize_interval_name(interval.name)

        # when an interval is pulled from the queue, pass it to the C backend
        result = _nfer.add(norm_name, interval.begin, interval.end, interval.data)
        # then try sending any generated intervals to the callbacks
        if result is not None:
            for (name, begin, end, data) in result:
                # denormalize the names that come back from nfer
                i = Interval(denormalize_interval_name(name), begin, end, data)
                if name in callbacks.keys():
                    for callback in callbacks[name]:
                        callback(i)
                # also call any global callbacks
                for callback in globcallbacks:
                    callback(i)


def monitor(what, callback=None):
    if isinstance(what, Interval):
        monitor(what.name, callback)

    elif isinstance(what, Rule):
        monitor(what.result, callback)
        # add the rule to the C backend after adding the callback
        # here, the rule can only be monitored once
        if not what.monitored:
            what.monitored = True
            _nfer.scan(what.to_rule_syntax())
            # if this is the first rule added to the C backend, start the monitor daemon
            if not monitor.spec_instantiated:
                # make a thread for the daemon
                nfer_thread = Thread(name='nfer.monitor', target=nfer_daemon, daemon=True)
                nfer_thread.start()
                # we have instantiated a spec in the C API
                monitor.spec_instantiated = True

    elif isinstance(what, str):
        # the string is the name of an interval to monitor
        if callback is not None:
            if what not in callbacks:
                callbacks[what] = []
            callbacks[what].append(callback)

monitor.spec_instantiated = False
